import altair as alt
import pandas as pd


def alt_line_chart(source: pd.DataFrame, *, x: str, y: str) -> alt.Chart:
    """Generate an altair line chart from a dataframe.

    The dataframe must be generated using `duqtools.ids.get_ids_dataframe` (or
    have the same format).

    Parameters
    ----------
    source : pd.DataFrame
        Input dataframe
    x : str
        X-value to plot, corresponds to a column in the source data
    y : str
        Y-value to plot, corresponds to a column in the source data

    Returns
    -------
    alt.Chart
        Return an altair chart.
    """
    slider = alt.binding_range(min=0, max=source['tstep'].max(), step=1)
    select_step = alt.selection_single(name='tstep',
                                       fields=['tstep'],
                                       bind=slider,
                                       init={'tstep': 0})

    chart = alt.Chart(source).mark_line().encode(
        x=f'{x}:Q',
        y=f'{y}:Q',
        color=alt.Color('run:N'),
        tooltip='run',
    ).add_selection(select_step).transform_filter(select_step).interactive()

    return chart


def alt_errorband_chart(source: pd.DataFrame, *, x: str, y: str) -> alt.Chart:
    """Generate an altair errorband plot from a dataframe.

    The dataframe must be generated using `duqtools.ids.get_ids_dataframe` (or
    have the same format).

    Parameters
    ----------
    source : pd.DataFrame
        Input dataframe
    x : str
        X-value to plot, corresponds to a column in the source data
    y : str
        Y-value to plot, corresponds to a column in the source data

    Returns
    -------
    alt.Chart
        Return an altair chart.
    """
    slider = alt.binding_range(min=0, max=source['tstep'].max(), step=1)
    select_step = alt.selection_single(name='tstep',
                                       fields=['tstep'],
                                       bind=slider,
                                       init={'tstep': 0})

    line = alt.Chart(source).mark_line().encode(
        x=f'{x}:Q',
        y=f'mean({y}):Q',
        color=alt.Color('tstep:N'),
    ).add_selection(select_step).transform_filter(select_step).interactive()

    # altair-viz.github.io/user_guide/generated/core/altair.ErrorBandDef
    band = alt.Chart(source).mark_errorband(
        extent='stdev', interpolate='linear').encode(
            x=f'{x}:Q',
            y=f'{y}:Q',
            color=alt.Color('tstep:N'),
        ).add_selection(select_step).transform_filter(
            select_step).interactive()

    chart = line + band

    return chart



from __future__ import annotations

import logging
from pathlib import Path

import click

from ._plot_utils import alt_line_chart
from .ids import ImasHandle, get_ids_dataframe
from .utils import read_imas_handles_from_file

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def plot(*, x_val, y_vals, imas_paths, input_files, dry_run, **kwargs):
    """Show subroutine to create plots from IDS data."""
    handles = {}

    for n, imas_path in enumerate(imas_paths):
        handle = ImasHandle.from_string(imas_path)
        handles[n] = handle

    for input_file in input_files:
        handles.update(read_imas_handles_from_file(input_file))

    if len(handles) == 0:
        raise SystemExit('No data to show.')

    source = get_ids_dataframe(handles, keys=(x_val, *y_vals))

    click.echo('You can now view your plot in your browser:')
    click.echo('')

    for n, y_val in enumerate(y_vals):
        chart = alt_line_chart(source, x=x_val, y=y_val)

        outfile = Path(f'chart_{n}.html')
        click.echo(f'  {x_val} vs. {y_val}:')
        click.secho(f'    file:///{outfile.absolute()}', bold=True)
        click.echo('')

        if not dry_run:
            chart.save(outfile, scale_factor=2.0)



from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Dict

from duqtools.schema.runs import Runs

from ._types import PathLike

if TYPE_CHECKING:
    from .ids import ImasHandle


@contextmanager
def work_directory(path: PathLike):
    """Changes working directory and returns to previous on exit.

    Parameters
    ----------
    path : PathLike
        Temporarily change to this directory.
    """
    prev_cwd = Path.cwd().resolve()
    try:
        os.chdir(path)
        yield
    finally:  # In any case, no matter what happens, go back eventually
        os.chdir(prev_cwd)


def read_imas_handles_from_file(inp: PathLike, ) -> Dict[str, ImasHandle]:
    """Read a collection of imas paths from a file.

    Input can be a `Runs.yaml` file `data.csv` file.

    The CSV file must have contain at least 5 columns, including: `user`,
    `db`, `shot`, and `run`. The first column is used as the index.
    The index can be any string or number, as long as it uniquely
    identifies the row.

    Parameters
    ----------
    inp : PathLike
        Name of the file to read.
    as_dataframe : bool, optional
        Return imas handles in a dataframe instead.

    Returns
    -------
    Union[Dict[str, ImasHandle], 'pd.DataFrame']
        Returns a dict with the Imas handles.

    Raises
    ------
    ValueError
        When the file cannot be opened.
    """
    import csv

    from .ids import ImasHandle

    inp = Path(inp)

    if inp.suffix == '.csv':
        handles = {}
        with open(inp) as f:
            reader = csv.DictReader(f)

            index_col = reader.fieldnames[0]  # type: ignore

            for row in reader:
                index = row.pop(index_col)
                handles[index] = ImasHandle(**row)

    elif inp.name == 'runs.yaml':
        runs = Runs.parse_file(inp)
        handles = {
            str(run.dirname): ImasHandle.parse_obj(run.data_out)
            for run in runs
        }

    else:
        raise ValueError(f'Cannot open file: {inp}')

    return handles



from __future__ import annotations

from ..schema.cli import ConfigModel


class Config(ConfigModel):
    """Config class containing all configs, is a singleton and can be used with
    import duqtools.config.Config as Cfg Cfg().<variable you want>"""

    _instance = None

    def __new__(cls, *args, **kwargs):
        # Make it a singleton
        if not Config._instance:
            Config._instance = object.__new__(cls)
        return Config._instance


cfg = Config()



from __future__ import annotations

import logging
from functools import singledispatch

import numpy as np

from ..schema import BaseModel, IDSOperation
from ._mapping import IDSMapping

logger = logging.getLogger(__name__)


@singledispatch
def apply_model(model: BaseModel, ids_mapping: IDSMapping) -> None:
    """Apply operation in model to IDS. Data are modified in-place.

    Parameters
    ----------
    model
        The model describes the operation to apply to the data.
    ids_mapping : IDSMapping
        Core profiles IDSMapping, data to apply operation to.
        Must contain the IDS path.

    Raises
    ------
    NotImplementedError
        When the model is unknown
    """

    raise NotImplementedError(f'Unknown model: {model}')


@apply_model.register
def _(model: IDSOperation, ids_mapping: IDSMapping) -> None:

    npfunc = getattr(np, model.operator)

    profile = ids_mapping.flat_fields[model.ids]

    if model.scale_to_error:
        sigma_key = model.ids + model._upper_suffix

        if model.value < 0:
            lower_key = model.ids + model._lower_suffix
            if lower_key in ids_mapping.flat_fields:
                sigma_key = lower_key

        sigma_bound = ids_mapping.flat_fields[sigma_key]
        sigma = abs(sigma_bound - profile)

        value = sigma * model.value
    else:
        value = model.value

    logger.info('Apply %s', model)

    logger.debug('data range before: %s - %s', profile.min(), profile.max())
    npfunc(profile, value, out=profile)
    logger.debug('data range after: %s - %s', profile.min(), profile.max())



from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple, Union

import numpy as np

from ._constants import TIME_COL, TSTEP_COL

if TYPE_CHECKING:
    import pandas as pd


def insert_re_caret_dollar(string: str) -> str:
    """Insert regex start (^) / end ($) of line matching characters."""
    if not string.startswith('^'):
        string = f'^{string}'
    if not string.endswith('$'):
        string = f'{string}$'
    return string


def recursive_defaultdict():
    """Recursive defaultdict."""
    return defaultdict(recursive_defaultdict)


def defaultdict_to_dict(ddict: defaultdict):
    """defaultdict_to_dict, turns a nested defaultdict into a nested dict.

    Parameters
    ----------
    ddict : defaultdict
        ddict
    """

    # Convert members to dict first
    for k, v in ddict.items():
        if isinstance(v, defaultdict):
            ddict[k] = defaultdict_to_dict(v)
    # Finally convert self to dict
    return dict(ddict)


class IDSMapping(Mapping):

    def __init__(self, ids, exclude_empty: bool = True):
        self._ids = ids
        self.exclude_empty = exclude_empty

        # All fields in the core profile in a single dict
        self.flat_fields: dict = {}

        # All fields, in the core profile in a nested dict
        self.fields: dict = defaultdict(recursive_defaultdict)

        self.dive(ids, [])
        self.fields = defaultdict_to_dict(self.fields)

    def __repr__(self):
        s = f'{self.__class__.__name__}(\n'
        for key in self.fields.keys():
            s += f'  {key} = ...\n'
        s += ')\n'

        return s

    def __getitem__(self, key: str):
        try:
            return self.flat_fields[key]
        except KeyError:
            pass
        try:
            return self.fields[key]
        except KeyError:
            raise

    def __iter__(self):
        yield from self.fields

    def __len__(self):
        return len(self.fields)

    def dive(self, val, path: list):
        """Recursively store the important bits of the imas structure in dicts.

        Parameters
        ----------
        val :
            Current nested object being evaluated
        path : List
            Current path
        """

        if isinstance(val, str):
            return

        if hasattr(val,
                   '__getitem__') and not isinstance(val,
                                                     (np.ndarray, np.generic)):
            for i in range(len(val)):
                item = val[i]
                self.dive(item, path + [str(i)])
            return

        if hasattr(val, '__dict__'):
            for key, item in val.__dict__.items():
                self.dive(item, path + [key])
            return

        if not isinstance(val, (np.ndarray, np.generic)):
            return

        if self.exclude_empty and val.size == 0:
            return

        # We made it here, the value can be stored
        self.flat_fields['/'.join(path)] = val
        cur = self.fields
        for item in path[:-1]:
            cur = cur[item]
        cur[path[-1]] = val

    def findall(self, pattern: str) -> Dict[str, Any]:
        """Find keys matching regex pattern.

        Parameters
        ----------
        pattern : str
            Regex pattern

        Returns
        -------
        dict
            New dict with all matching key/value pairs.
        """
        pattern = insert_re_caret_dollar(pattern)

        pat = re.compile(pattern)

        return {
            key: self.flat_fields[key]
            for key in self.flat_fields if pat.match(key)
        }

    def find_by_group(self, pattern: str) -> Dict[Union[tuple, str], Any]:
        """Find keys matching regex pattern by group.

        The dict key is defined by `match.groups()`.
        Dict entries will be overwritten if the groups are not unique.

        Parameters
        ----------
        pattern : str
            Regex pattern (must contain groups)

        Returns
        -------
        dict
            New dict with all matching key/value pairs.
        """
        pattern = insert_re_caret_dollar(pattern)

        pat = re.compile(pattern)

        new = {}
        for key in self.flat_fields:
            m = pat.match(key)
            if m:
                groups = m.groups()
                idx = groups[0] if len(groups) == 1 else groups
                new[idx] = self.flat_fields[key]

        return new

    def find_by_index(self, pattern: str) -> Dict[str, Dict[int, np.ndarray]]:
        """Find keys matching regex pattern using time index.

        Must include $i, which is a special character that matches
        an integer (`\\d+`)

        i.e. `ids.find_by_index('profiles_1d/$i/zeff.*')`
        returns a dict with `zeff` and error attributes.

        Parameters`
        ----------
        pattern : str
            Regex pattern, must include a group matching a digit.

        Returns
        -------
        dict
            New dict with all matching key/value pairs.
        """
        idx_str = '$i'

        if idx_str not in pattern:
            raise ValueError(f'Pattern must include {idx_str} to match index.')

        pattern = insert_re_caret_dollar(pattern)

        pattern = pattern.replace(idx_str, r'(?P<idx>\d+)')
        pat = re.compile(pattern)

        new_dict: Dict[str, Dict[int, np.ndarray]] = defaultdict(dict)

        for key in self.flat_fields:
            m = pat.match(key)

            if m:
                si, sj = m.span('idx')
                new_key = key[:si] + idx_str + key[sj:]

                idx = int(m.group('idx'))
                new_dict[new_key][idx] = self.flat_fields[key]

        return new_dict

    def to_dataframe(self,
                     *variables: str,
                     prefix: str = 'profiles_1d',
                     time_steps: Sequence[int] = None) -> pd.DataFrame:
        """Return long format dataframe for given variables.

        Search string:
        `{prefix}/{time_step}/{variable}`

        Parameters
        ----------
        *variables : str
            Keys to extract, i.e. `zeff`, `grid/rho_tor`
        prefix : str, optional
            First part of the data path
        time_steps : Sequence[int], optional
            List or array of integer time steps to extract.
            Defaults to all time steps.

        Returns
        -------
        df : pd.DataFrame
            Contains a column for the time step and each of the variables.
        """
        import pandas as pd
        columns, arr = self.to_numpy(*variables,
                                     prefix=prefix,
                                     time_steps=time_steps)

        df = pd.DataFrame(arr, columns=columns)
        df[TSTEP_COL] = df[TSTEP_COL].astype(int)

        return df

    def to_numpy(
        self,
        *variables: str,
        prefix: str = 'profiles_1d',
        time_steps: Sequence[int] = None
    ) -> Tuple[Tuple[str, ...], np.ndarray]:
        """Return numpy array containing data for given variables.

        Search string:
        `{prefix}/{time_step}/{variable}`

        Parameters
        ----------
        *variables : str
            Keys to extract, i.e. `zeff`, `grid/rho_tor`
        prefix : str, optional
            First part of the data path
        time_steps : Sequence[int], optional
            List or array of integer time steps to extract.
            Defaults to all time steps.

        Returns
        -------
        columns, array : Tuple[Tuple[str], np.ndarray]
            Numpy array with a column for the time step and each of the
            variables.
        """
        points_per_var = len(self.flat_fields[f'{prefix}/0/{variables[0]}'])

        if not time_steps:
            n_time_steps = len(self[TIME_COL])
            time_steps = range(n_time_steps)
        else:
            n_time_steps = len(time_steps)

        columns = (TSTEP_COL, TIME_COL, *variables)
        n_vars = len(columns)

        arr = np.empty((n_time_steps * points_per_var, n_vars))

        timestamps = self[TIME_COL]

        for t in time_steps:
            for j, variable in enumerate(variables):
                flat_variable = f'{prefix}/{t}/{variable}'

                i_begin = t * points_per_var
                i_end = i_begin + points_per_var

                arr[i_begin:i_end, 0] = t
                arr[i_begin:i_end, 1] = timestamps[t]
                arr[i_begin:i_end, j + 2] = self.flat_fields[flat_variable]

        return columns, arr



import logging
from typing import Iterable

from duqtools.config import cfg

from .ids import IDSMapping, ImasHandle, apply_model
from .matrix_samplers import get_matrix_sampler
from .models import WorkDirectory
from .operations import add_to_op_queue, confirm_operations, op_queue
from .schema.runs import Runs
from .system import get_system

logger = logging.getLogger(__name__)

RUN_PREFIX = 'run_'


def fail_if_locations_exist(locations: Iterable[ImasHandle]):
    """Check IDS coordinates and raise if any exist."""
    any_exists = False
    for location in locations:
        if location.exists():
            logger.info('Target %s already exists', location)
            any_exists = True
    if any_exists:
        raise IOError(
            'Found existing target location(s), use `duqtools clean` to '
            'remove or `--force` to override.')


@add_to_op_queue('Setting inital condition of', '{target_in}')
def apply_combination(target_in: ImasHandle, combination) -> None:
    core_profiles = target_in.get('core_profiles')
    ids_mapping = IDSMapping(core_profiles)

    for model in combination:
        apply_model(model, ids_mapping)

        logger.info('Writing data entry: %s', target_in)
        with target_in.open() as data_entry_target:
            core_profiles.put(db_entry=data_entry_target)


@add_to_op_queue('Writing out', '{workspace.runs_yaml}')
def write_runs_file(runs: list, workspace) -> None:

    runs = Runs.parse_obj(runs)
    with open(workspace.runs_yaml, 'w') as f:
        runs.yaml(stream=f)


@confirm_operations
def create(*, force, **kwargs):
    """Create input for jetto and IDS data structures.

    Parameters
    ----------
    force : bool
        Override protection if data and directories already exist.
    **kwargs
        Unused.
    """
    options = cfg.create
    if not options:
        logger.warning('No create options specified.')
        return

    workspace = WorkDirectory.parse_obj(cfg.workspace)

    template_drc = options.template
    dimensions = options.dimensions
    matrix_sampler = get_matrix_sampler(options.sampler.method)

    system = get_system()

    if not options.template_data:
        source = system.imas_from_path(template_drc)
    else:
        source = ImasHandle.parse_obj(options.template_data)

    logger.info('Source data: %s', source)

    matrix = tuple(dim.expand() for dim in dimensions)
    combinations = matrix_sampler(*matrix, **dict(options.sampler))

    if not force:
        if workspace.runs_yaml.exists():
            raise IOError(
                'Directory is not empty, use `duqtools clean` to clear or '
                '`--force` to override.')

        locations = (ImasHandle(db=options.data.db,
                                shot=source.shot,
                                run=options.data.run_in_start_at + i)
                     for i in range(len(combinations)))

        fail_if_locations_exist(locations)

    runs = []

    for i, combination in enumerate(combinations):
        run_name = f'{RUN_PREFIX}{i:04d}'
        run_drc = workspace.cwd / run_name

        op_queue.add(action=run_drc.mkdir,
                     kwargs={
                         'parents': True,
                         'exist_ok': force
                     },
                     description='Create folder',
                     extra_description=f'{run_drc}')

        target_in = ImasHandle(db=options.data.db,
                               shot=source.shot,
                               run=options.data.run_in_start_at + i)
        target_out = ImasHandle(db=options.data.db,
                                shot=source.shot,
                                run=options.data.run_out_start_at + i)

        source.copy_ids_entry_to(target_in)

        apply_combination(target_in, combination)

        system.copy_from_template(template_drc, run_drc)
        system.write_batchfile(workspace, run_name)

        system.update_imas_locations(run=run_drc,
                                     inp=target_in,
                                     out=target_out)

        runs.append({
            'dirname': run_name,
            'data_in': target_in,
            'data_out': target_out,
            'operations': combination
        })

    write_runs_file(runs, workspace)



import logging
from typing import Sequence

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from ._constants import RUN_COL, TIME_COL, TSTEP_COL

logger = logging.getLogger(__name__)


def rebase_on_ids(source: pd.DataFrame,
                  *,
                  base_col: str,
                  value_cols: Sequence[str],
                  new_base: np.ndarray = None) -> pd.DataFrame:
    """Rebase data on new ids basis using interpolation.

    This operation makes sure that all data on the x-axis are the same for
    each run and time step.

    Uses [scipy.interpolate.interp1d][].

    Parameters
    ----------
    source : pd.DataFrame
        Input data, contains the columns 'run', 'tstep' and any number of
        ids columns.
    base_col : str
        This defines the base ids column that the new base belongs to.
        In other words, this is the `x` column in the interpolation.
    value_cols : Sequence[str]
        The data in these ids columns will be interpolated.
        In other words, these are the `y` columns in the interpolation.
        IDS columns not defined by base_col and value_cols will be omitted
        from the output.
    new_base : np.ndarray, optional
        Numpy array with the new base values for the given base column.
        If not defined, use the data in the base column of the first time
        step of the first run as the basis.

    Returns
    -------
    pd.DataFrame
        For the returned dataframe, for each run and time step,
        the values in the base column will be the same.
    """
    if new_base is None:
        first_run = source.iloc[0].run
        idx = (source[RUN_COL] == first_run) & (source[TSTEP_COL] == 0)
        new_base = source[idx][base_col]
        logger.debug('Rebase ids on %s, using %s from %d to %d with %d steps',
                     first_run, base_col, new_base.min(), new_base.max(),
                     len(new_base))

    def refit(gb: pd.DataFrame) -> pd.DataFrame:
        new_values = []

        for value_col in value_cols:
            f = interp1d(gb[base_col],
                         gb[value_col],
                         fill_value='extrapolate',
                         bounds_error=False)
            new_values.append(f(new_base))

        df = pd.DataFrame((new_base, *new_values),
                          index=[base_col, *value_cols]).T

        df[TIME_COL] = gb[TIME_COL].iloc[0]
        return df

    grouped = source.groupby([RUN_COL, TSTEP_COL])

    out = grouped.apply(refit).reset_index(
        (RUN_COL, TSTEP_COL)).reset_index(drop=True)

    out = out[[RUN_COL, TSTEP_COL, TIME_COL, base_col, *value_cols]]
    return out


def rebase_on_time(source: pd.DataFrame,
                   *,
                   cols: Sequence[str],
                   new_base: np.ndarray = None) -> pd.DataFrame:
    """Rebase data on new time basis using interpolation.

    This operation makes sure that each run has the same time steps.

    Uses [scipy.interpolate.interp1d][].

    Parameters
    ----------
    source : pd.DataFrame
        Input data, contains the columns 'run', 'tstep' and any number of
        ids columns.
    cols : Sequence[str]
        This defines the columns that should be rebased.
        IDS columns not defined will be omitted from the output.
    new_base : np.ndarray, optional
        Numpy array with the new base values for the time steps.
        If not defined, use the time steps in the first run of the
        source data.

    Returns
    -------
    pd.DataFrame
        For the returned dataframe, for each run the time steps will
        be the same.
    """
    if new_base is None:
        first_run = source.iloc[0].run
        new_base = source[source[RUN_COL] == first_run][TIME_COL].unique()
        logger.debug('Rebase time on %s, from %d to %d with %d steps',
                     first_run, new_base.min(), new_base.max(), len(new_base))

    cols = list(cols)

    n_cols = len(cols)
    n_time_new = len(new_base)

    def refit(gb: pd.DataFrame) -> pd.DataFrame:
        time = gb[TIME_COL].unique()
        values = np.array(gb[cols])

        n_times = len(time)
        n_vals = int(len(values) / n_times)

        values = values.reshape(n_times, n_vals, n_cols).T

        f = interp1d(time,
                     values,
                     fill_value='extrapolate',
                     bounds_error=False)

        values_new = f(new_base)
        values_new = values_new.T.reshape(n_time_new * n_vals, n_cols)

        time_new = np.repeat(new_base, n_vals).reshape(-1, 1)

        tstep_new = np.repeat(np.arange(n_time_new), n_vals).reshape(-1, 1)

        arr = np.hstack((tstep_new, time_new, values_new))

        out = pd.DataFrame(arr, columns=[TSTEP_COL, TIME_COL, *cols])
        out[TSTEP_COL] = out[TSTEP_COL].astype(np.int64)
        return out

    grouped = source.groupby([RUN_COL])

    out = grouped.apply(refit).reset_index(RUN_COL).reset_index(drop=True)
    return out



# https://github.com/CarbonCollective/fusion-dUQtools/issues/27

import logging

logger = logging.getLogger(__name__)

try:
    import imas
    from imas import imasdef
except (ModuleNotFoundError, ImportError):
    from unittest.mock import MagicMock as Mock
    imas = Mock()
    imasdef = Mock()

    logger.warning('Could not import IMAS:', exc_info=True)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Sequence

from ._get_ids_tree import get_ids_tree
from ._handle import ImasHandle

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import pandas as pd


def _get_ids_run_dataframe(handle: ImasHandle, *, keys,
                           **kwargs) -> pd.DataFrame:
    """Get data for single run."""
    logger.info('Getting data for %s', handle)
    profile = get_ids_tree(handle, exclude_empty=True)
    return profile.to_dataframe(*keys, **kwargs)


def get_ids_dataframe(handles: Dict[str, ImasHandle], *, keys: Sequence[str],
                      **kwargs) -> pd.DataFrame:
    """Read a dict of IMAS handles into a structured pandas dataframe.

    The returned dataframe will have the columns:

        `run`, `tsteps`, `times`, `<ids col 1>, `<ids_col_2>`, ...

    Where `tstep` corresponds to the time index, and `times` to the
    actual times.

    Parameters
    ----------
    handles : Dict[str, ImasHandle]
        Dict with IMAS handles. The key is used as the 'run' name in
        the dataframe.
    keys : Sequence[str]
        IDS values to extract. These will be used as columns in the
        data frame.
    **kwargs
        These keyword parameters are passed to
        `duqtools.ids.IDSMapping.to_dataframe`.

    Returns
    -------
    pd.DataFrame
        Description
    """
    import pandas as pd

    runs_data = {
        str(name): _get_ids_run_dataframe(handle, keys=keys, **kwargs)
        for name, handle in handles.items()
    }

    return pd.concat(runs_data,
                     names=('run',
                            'index')).reset_index('run').reset_index(drop=True)



import logging
import shutil
import stat
from pathlib import Path

logger = logging.getLogger(__name__)


def copy_files(source_drc: Path, target_drc: Path):
    """Copy files for jetto run to destination directory.

    Parameters
    ----------
    source_drc : Path
        Source (template) directory.
    target_drc : Path
        Target directory.
    """
    for filename in (
            # '.llcmd',
            'jetto.in',
            'rjettov',
            'utils_jetto',
            'jetto.ex',
            'jetto.sin',
            'jetto.sgrid',
            'jetto.jset',
    ):
        src = source_drc / filename
        dst = target_drc / filename
        shutil.copyfile(src, dst)

    for filename in ('rjettov', 'utils_jetto'):
        path = target_drc / filename
        path.chmod(path.stat().st_mode | stat.S_IEXEC)

    logger.debug('copied files to %s', target_drc)



"""Functions to interface with `jetto.in` namelists."""

import f90nml

from .._types import PathLike


def read_namelist(path: PathLike) -> dict:
    """Read fortran namelist (i.e. `jetto.in`).

    Parameters
    ----------
    path : PathLike
        Path to namelist

    Returns
    -------
    namelist : dict
        Returns parameters in namelist as dict
    """
    return f90nml.read(path).todict()


def write_namelist(path: PathLike, namelist: dict):
    """Write dictionary fortran namelist (i.e. `jetto.in`).

    Parameters
    ----------
    path : PathLike
        Path to namelist
    namelist : dict
        Fortran namelist in dictionary format
    """
    nml = f90nml.Namelist(**namelist)
    nml.write(path)


def patch_namelist(path: PathLike, patch: dict, out: PathLike):
    """Patch parameters in namelist.

    i.e. in the case of `jetto.in`, this preserves
    the comments in the header, which are required for
    `rjettov`.

    Parameters
    ----------
    path : str
        Path to original namelist
    patch : dict
        Dictionary with variables to patch / add
    out : PathLike
        Path to write patched `jetto.in` file.

    Returns
    -------
    patched_nml : dict
        Return patched namelist as dictionary
    """
    patched_nml = f90nml.patch(path, patch, out)
    return patched_nml.todict()



import logging
from datetime import datetime
from sys import stderr, stdout

import click
import coverage

from .config import cfg
from .operations import op_queue

logger = logging.getLogger(__name__)
coverage.process_startup()


def config_option(f):

    def callback(ctx, param, config):
        if ctx.command.name != 'init':
            cfg.parse_file(config)

        return config

    return click.option('-c',
                        '--config',
                        default='duqtools.yaml',
                        help='Path to config.',
                        callback=callback)(f)


def debug_option(f):

    def callback(ctx, param, debug):
        if debug:
            logging.getLogger().setLevel(logging.DEBUG)

        return debug

    return click.option('--debug',
                        is_flag=True,
                        help='Enable debug print statements.',
                        callback=callback)(f)


def dry_run_option(f):

    def callback(ctx, param, dry_run):
        if dry_run:
            logger.info('--dry-run enabled')
            op_queue.dry_run = True
        else:
            op_queue.dry_run = False

        return dry_run

    return click.option('--dry-run',
                        is_flag=True,
                        help='Execute without any side-effects.',
                        callback=callback)(f)


def yes_option(f):

    def callback(ctx, param, yes):
        if yes:
            logger.info('--yes enabled')
            op_queue.yes = True
        else:
            op_queue.yes = False
        return yes

    return click.option('--yes',
                        is_flag=True,
                        help='Answer yes to questions automatically.',
                        callback=callback)(f)


def logfile_option(f):

    def callback(ctx, param, logfile):
        streams = {'stdout': stdout, 'stderr': stderr}

        logger.info(f'logging to {logfile}')
        logging.getLogger().handlers = []

        if logfile in streams.keys():
            logging.basicConfig(stream=streams[logfile], level=logging.INFO)
        else:
            logging.basicConfig(filename=logfile, level=logging.INFO)

        logger.info('')
        logger.info(
            'Duqtools starting at '
            f'{datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S %z")}')
        logger.info('------------------------------------------------')
        logger.info('')

        return logfile

    return click.option('--logfile',
                        '-l',
                        is_flag=False,
                        default='duqtools.log',
                        help='where to send the logfile,'
                        ' the special values stderr/stdout'
                        ' will send it there respectively.',
                        callback=callback)(f)


def common_options(func):
    for wrapper in (logfile_option, debug_option, config_option,
                    dry_run_option, yes_option):
        # config_option MUST BE BEFORE dry_run_option
        # logfile_option must be before debug_option
        func = wrapper(func)
    return func


@click.group()
def cli(**kwargs):
    """For more information, check out the documentation:

    https://duqtools.readthedocs.io
    """
    pass


@cli.command('init')
@common_options
@click.option('--full',
              is_flag=True,
              help='Create a config file with all possible config values.')
@click.option('--force', is_flag=True, help='Overwrite existing config.')
def cli_init(**kwargs):
    """Create a default config file."""
    from .init import init
    init(**kwargs)


@cli.command('create')
@common_options
@click.option('--force',
              is_flag=True,
              help='Overwrite existing run directories and IDS data.')
def cli_create(**kwargs):
    """Create the UQ run files."""
    from .create import create
    create(**kwargs)


@cli.command('submit')
@common_options
@click.option('--force', is_flag=True, help='Re-submit running jobs.')
def cli_submit(**kwargs):
    """Submit the UQ runs."""
    from .submit import submit
    submit(**kwargs)


@cli.command('status')
@common_options
@click.option('--detailed', is_flag=True, help='Detailed info on progress')
@click.option('--progress', is_flag=True, help='Fancy progress bar')
def cli_status(**kwargs):
    """Print the status of the UQ runs."""
    from .status import status
    status(**kwargs)


@cli.command('plot')
@click.option('-x', 'x_val', type=str, help='IDS of the x value')
@click.option('-y',
              'y_vals',
              type=str,
              help='IDS of the y value',
              multiple=True)
@click.option('-m',
              '--imas',
              'imas_paths',
              type=str,
              help='IMAS path formatted as <user>/<db>/<shot>/<number>.',
              multiple=True)
@click.option('-i',
              '--input',
              'input_files',
              type=str,
              help='Input file, i.e. `data.csv` or `runs.yaml`',
              multiple=True)
@common_options
def cli_plot(**kwargs):
    """Plot some IDS data."""
    from .plot import plot
    plot(**kwargs)


@cli.command('clean')
@common_options
@click.option('--out', is_flag=True, help='Remove output data.')
@click.option('--force',
              is_flag=True,
              help='Overwrite backup file if necessary.')
def cli_clean(**kwargs):
    """Delete generated IDS data and the run dir."""
    from .cleanup import cleanup
    cleanup(**kwargs)


@cli.command('dash')
@common_options
def cli_dash(**kwargs):
    """Open dashboard for evaluating IDS data."""
    from .dash import dash
    dash(**kwargs)


if __name__ == '__main__':
    cli()



import logging

from ._dimensions import apply_model
from ._get_ids_tree import get_ids_tree
from ._handle import ImasHandle
from ._io import get_ids_dataframe
from ._mapping import IDSMapping
from ._rebase import rebase_on_ids, rebase_on_time

logger = logging.getLogger(__name__)

__all__ = [
    'apply_model',
    'get_ids_dataframe',
    'get_ids_tree',
    'IDSMapping',
    'ImasHandle',
    'rebase_on_ids',
    'rebase_on_time',
]



"""Function to create llcmd file."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from duqtools.models import WorkDirectory


def write_batchfile(workspace: WorkDirectory, run_name: str):
    """Write batchfile (`.llcmd`) to start jetto.

    Parameters
    ----------
    target_drc : Path
        Directory to place batch file into.
    """
    run_drc = workspace.cwd / run_name
    llcmd_path = run_drc / '.llcmd'

    full_path = workspace.cwd / run_name
    rjettov_path = full_path / 'rjettov'
    rel_path = workspace.subdir / run_name

    with open(llcmd_path, 'w') as f:
        f.write(f"""#!/bin/sh
#SBATCH -J jetto.{run_name}
#SBATCH -i /dev/null
#SBATCH -o ll.out
#SBATCH -e ll.err
#SBATCH -p gw

#SBATCH -N 1
#SBATCH -n 2
#SBATCH -t 24:00:00

cd {full_path}
{rjettov_path} -S -I -p -xmpi -x64 {rel_path} v210921_gateway_imas g2fkoech
""")



from __future__ import annotations

from typing import List

from pydantic import DirectoryPath, Field

from ._basemodel import BaseModel
from ._dimensions import IDSOperation
from ._imas import ImasBaseModel


class Run(BaseModel):
    dirname: DirectoryPath = Field(None, description='Directory of run')
    data_in: ImasBaseModel
    data_out: ImasBaseModel
    operations: List[IDSOperation]


class Runs(BaseModel):
    __root__: List[Run] = []

    def __iter__(self):
        yield from self.__root__

    def __getitem__(self, index: int):
        return self.__root__[index]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ..ids import ImasHandle

if TYPE_CHECKING:
    from ._jset import JettoSettings

logger = logging.getLogger(__name__)


def imas_from_jset_input(jset: JettoSettings) -> ImasHandle:
    """Get IMAS input location from jetto settings.

    Parameters
    ----------
    jset : JettoSettings
        Jetto settings.

    Returns
    -------
    destination : ImasHandle
        Returns the destination.
    """
    return ImasHandle(
        db=jset.machine_in,  # type: ignore
        user=jset.user_in,  # type: ignore
        run=jset.run_in,  # type: ignore
        shot=jset.shot_in)  # type: ignore


def imas_from_jset_output(jset: JettoSettings) -> ImasHandle:
    """Get IMAS output location from jetto settings.

    Parameters
    ----------
    jset : JettoSettings
        Jetto settings.

    Returns
    -------
    destination : ImasHandle
        Returns the destination.
    """
    return ImasHandle(
        db=jset.machine_out,  # type: ignore
        run=jset.run_out,  # type: ignore
        shot=jset.shot_out)  # type: ignore



from __future__ import annotations

import logging
from collections import deque
from inspect import signature
from typing import Callable, Optional

import click
from pydantic import Field

from .schema import BaseModel

logger = logging.getLogger(__name__)


class Operation(BaseModel):
    """Operation, simple class which has a callable action.

    Usually not called directly but used through Operations. has the
    following members:
    """

    description: str = Field(
        description='description of the operation to be done')
    action: Callable = Field(
        description='a function which can be executed when we '
        'decide to apply this operation')
    args: tuple = Field((),
                        description='positional arguments that have to be '
                        'passed to the action')
    kwargs: dict = Field(None,
                         description='keyword arguments that will be '
                         'passed to the action')

    def __init__(self,
                 action,
                 description,
                 extra_description: Optional[str] = None,
                 *args,
                 **kwargs):
        """__init__.

        Parameters
        ----------
        description: str
            description of the operation to be done
        extra_description: Optional[str]
            extra description, (not colorized)
        action: Callable
            function to be eventually evaluated
        *args: tuple
            positional arguments to action
        **kwargs: dict
            keyword arguments to action
        """

        if extra_description:
            description = click.style(description,
                                      fg='green') + ' : ' + extra_description
        super().__init__(action=action,
                         description=description,
                         *args,
                         **kwargs)
        if not self.kwargs:
            self.kwargs = {}

    def __call__(self) -> Operation:
        """Execute the action with the args and kwargs.

        Returns
        -------
        Operation
            The operation that was executed
        """
        logger.debug(self.description)
        self.action(*self.args, **self.kwargs)
        return self


class Operations(deque):
    """Operations Queue which keeps track of all the operations that need to be
    done.

    It's basically dask_delayed, but custom made and a few drawbacks:
    The return value from an action is eventually discarded,
    communication between queue items is possible through references, or
    global values, but not really recommended, and no guidance for this
    is provided
    """

    _instance = None
    yes = False  # Apply operations without prompt
    dry_run = False  # Never apply any operations (do not even ask)
    logger = logging.getLogger('operations')

    def __init__(self):
        super().__init__()
        stream = logging.StreamHandler()
        stream.setFormatter(logging.Formatter('%(message)s'))
        self.logger.addHandler(stream)

    def __new__(cls, *args, **kwargs):
        # Make it a singleton
        if not Operations._instance:
            Operations._instance = super().__new__(cls)
        return Operations._instance

    def add(self, *args, **kwargs) -> None:
        """convenience Operation wrapper around put.

        ```python
        from duqtools.operations import add_to_op_queue.

        op_queue.add(print, args=('Hello World,),
                description="Function that prints hello world")
        ```
        """

        self.append(Operation(*args, **kwargs))

    def put(self, item: Operation) -> None:
        """synonym for append."""
        self.append(item)

    def append(self, item: Operation) -> None:  # type: ignore
        """Restrict our diet to Operation objects only."""

        super().append(item)

    def apply(self) -> Operation:
        """Apply the next operation in the queue and remove it."""

        operation = self.popleft()
        operation()
        return operation

    def apply_all(self) -> None:
        """Apply all queued operations and empty the queue.

        and show a fancy progress bar while applying
        """
        from tqdm import tqdm
        logger.info(click.style('Applying Operations', fg='red', bold=True))
        with tqdm(total=len(self), position=1) as pbar:
            pbar.set_description('Applying operations')
            with tqdm(iterable=False, bar_format='{desc}', position=0) as dbar:
                while len(self) != 0:
                    op = self.popleft()
                    op()
                    dbar.set_description(op.description)
                    logger.info(op.description)
                    pbar.update()

    def confirm_apply_all(self) -> bool:
        """First asks the user if he wants to apply everything.

        Returns
        -------
        bool: did we apply everything or not
        """

        # To print the descriptions we need to get them
        self.logger.info('')
        self.logger.info(
            click.style('Operations in the Queue:', fg='red', bold=True))
        self.logger.info(click.style('========================', fg='red'))
        for op in self:
            self.logger.info('- ' + op.description)

        if self.dry_run:
            self.logger.info('Dry run enabled, not applying op_queue')
            return False

        ans = self.yes or click.confirm(
            'Do you want to apply all these operations?', default=False)

        if ans:
            self.apply_all()
        return ans


op_queue = Operations()


def confirm_operations(func):
    """Decorator which confirms and applies queued operations after the
    function.

    ```python
    from duqtools.operations import confirm_operations, op_queue

    @confirm_operations
    def complicated_stuff()
        op_queue.add(print, args=('Hello World,),
                description="Function that prints hello world")
        op_queue.add(print, args=('Hello World again,),
                description="Function that prints hello world, again")
    ```
    """

    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        op_queue.confirm_apply_all()
        op_queue.clear()
        return ret

    return wrapper


def add_to_op_queue(op_desc: str, extra_desc: Optional[str] = None):
    """Decorator which adds the function call to the op_queue, instead of
    executing it directly, the string can be a format string and use the
    function arguments.

    ```python
    from duqtools.operations import add_to_op_queue, op_queue

    @add_to_op_queue("Printing hello world", "{name}")
    def print_hello_world(name):
        print(f"Hello World {name}")


    print_hello_world("Snoozy")

    op_queue.confirm_apply_all()
    ```
    """

    def op_queue_real(func):

        def wrapper(*args, **kwargs) -> None:
            # For the description format we must convert args to kwargs
            sig = signature(func)
            args_to_kw = dict(zip(sig.parameters, args))
            fkwargs = kwargs.copy()
            fkwargs.update(args_to_kw)
            extra_formatted = None
            if extra_desc:
                extra_formatted = extra_desc.format(**fkwargs)
            op_formatted = op_desc.format(**fkwargs)

            # add the function to the queue
            op_queue.add(action=func,
                         args=args,
                         kwargs=kwargs,
                         description=op_formatted,
                         extra_description=extra_formatted)

        return wrapper

    return op_queue_real



"""This module contains tools for interfacing with jetto runs."""

from ._copy import copy_files
from ._jset import JettoSettings, read_jset, write_jset
from ._llcmd import write_batchfile
from ._namelist import patch_namelist, read_namelist, write_namelist
from ._system import JettoSystem

__all__ = [
    'read_namelist',
    'patch_namelist',
    'write_namelist',
    'read_jset',
    'write_jset',
    'write_batchfile',
    'JettoSettings',
    'copy_files',
    'JettoSystem',
]



from pathlib import Path

from .config import cfg
from .ids import ImasHandle
from .jetto import JettoSystem
from .models import AbstractSystem, WorkDirectory


class DummySystem(AbstractSystem):
    """This is a dummy system that implements the basic interfaces.

    It exists for testing purposes in absence of actual modelling
    software.
    """

    @staticmethod
    def write_batchfile(workspace: WorkDirectory, run_name: str):
        pass

    @staticmethod
    def copy_from_template(source_drc: Path, target_drc: Path):
        pass

    @staticmethod
    def imas_from_path(template_drc: Path):
        return ImasHandle(db='', shot='-1', run='-1')

    @staticmethod
    def update_imas_locations(run: Path, inp, out):
        pass


def get_system():
    """get_system.

    Get the system to do operations with TODO make it a variable, not a
    function
    """
    if (cfg.system == 'jetto'):
        return JettoSystem
    elif (cfg.system == 'dummy'):
        return DummySystem
    else:
        raise NotImplementedError(f'system {cfg.system} is not implemented')



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ._mapping import IDSMapping

if TYPE_CHECKING:
    from .ids import ImasHandle

logger = logging.getLogger(__name__)


def get_ids_tree(imas_loc: ImasHandle,
                 key: str = 'core_profiles',
                 **kwargs) -> IDSMapping:
    """get the data as a simple ids (all values in memory, in a dict).

    Parameters
    ----------
    key : str, optional
        Name of profiles to open
    **kwargs
        These parameters are passed to initialize `IDSMapping`.

    Returns
    -------
    IDSMapping
    """
    return IDSMapping(imas_loc.get(key), **kwargs)



from __future__ import annotations

from typing import List, Tuple, Union

import numpy as np
from pydantic import Field, validator
from typing_extensions import Literal

from ._basemodel import BaseModel
from ._description_helpers import formatter as f


class IDSPathMixin(BaseModel):
    ids: str = Field('profiles_1d/0/t_i_average',
                     description=f("""
            IDS Path of the data to modify. `core_profiles` is implied.
            """))


class IDSOperatorMixin(BaseModel):
    operator: Literal['add', 'multiply', 'divide', 'power', 'subtract',
                      'floor_divide', 'mod',
                      'remainder'] = Field('multiply',
                                           description=f("""
        Which operator to apply to the data in combination with any of the
        given values below. This can be any of the basic numpy arithmetic
        operations. Available choices: `add`, `multiply`, `divide`, `power`,
        `subtract`, `floor_divide`, `mod`, and `remainder`. These directly map
        to the equivalent numpy functions, i.e. `add` -> `np.add`.
        """))
    scale_to_error: bool = Field(False,
                                 description=f("""
        If True, multiply value(s) by the error (sigma).

        With asymmetric errors (i.e. both lower/upper error are available),
        scale negative values to the lower error, and positive values to upper
        error.
        """))

    _upper_suffix: str = '_error_upper'
    _lower_suffix: str = '_error_lower'


class IDSOperation(IDSPathMixin, IDSOperatorMixin, BaseModel):
    """Apply arithmetic operation to IDS."""

    value: float = Field(description=f("""
        Values to use with operator on field to create sampling
        space."""))


class LinSpace(BaseModel):
    """Generated evenly spaced numbers over a specified interval.

    See the implementation of [numpy.linspace][] for more details.
    """
    start: float = Field(None, description='Start value of the sequence.')
    stop: float = Field(None, description='End value of the sequence.')
    num: int = Field(None, description='Number of samples to generate.')

    @property
    def values(self):
        """Convert to list."""
        # `val.item()` converts to native python types
        return [
            val.item() for val in np.linspace(self.start, self.stop, self.num)
        ]


class ARange(BaseModel):
    """Generate evenly spaced numbers within a given interval.

    See the implementation of [numpy.arange][] for more details.
    """

    start: float = Field(
        None, description='Start of the interval. Includes this value.')
    stop: float = Field(
        None, description='End of the interval. Excludes this interval.')
    step: float = Field(None, description='Spacing between values.')

    @property
    def values(self):
        """Convert to list."""
        # `val.item()` converts to native python types
        return [
            val.item() for val in np.arange(self.start, self.stop, self.step)
        ]


class IDSOperationDim(IDSPathMixin, IDSOperatorMixin, BaseModel):
    """Apply set of arithmetic operations to IDS.

    Takes the IDS data and subtracts, adds, multiplies, etc with each
    the given values.
    """

    values: Union[List[float], ARange, LinSpace, ] = Field([1.1, 1.2, 1.3],
                                                           description=f("""
            Values to use with operator on field to create sampling
            space."""))

    def expand(self, *args, **kwargs) -> Tuple[IDSOperation, ...]:
        """Expand list of values into operations with its components."""
        return tuple(
            IDSOperation(ids=self.ids,
                         operator=self.operator,
                         value=value,
                         scale_to_error=self.scale_to_error)
            for value in self.values)

    @validator('values')
    def convert_to_list(cls, v):
        if not isinstance(v, list):
            v = v.values
        return v



from getpass import getuser
from typing import List, Union

from pydantic import DirectoryPath, Field, validator
from typing_extensions import Literal

from ._basemodel import BaseModel
from ._description_helpers import formatter as f
from ._dimensions import IDSOperationDim
from ._imas import ImasBaseModel
from .data_location import DataLocation
from .matrix_samplers import (CartesianProduct, HaltonSampler, LHSSampler,
                              SobolSampler)
from .workdir import WorkDirectoryModel


class DeprecatedValueError(ValueError):
    ...


class CreateConfigModel(BaseModel):
    """The options of the `create` subcommand are stored in the `create` key in
    the config."""
    dimensions: List[Union[IDSOperationDim]] = Field([
        IDSOperationDim(ids='profiles_1d/0/t_i_average', ),
        IDSOperationDim(ids='profiles_1d/0/electrons/temperature')
    ],
                                                     description=f("""
        The `dimensions` specifies the dimensions of the matrix to sample
        from. Each dimension is a compound set of operations to apply.
        From this, a matrix all possible combinations is generated.
        Essentially, it generates the
        [Cartesian product](en.wikipedia.org/wiki/Cartesian_product)
        of all operations. By specifying a different `sampler`, a subset of
        this hypercube can be efficiently sampled.
        """))

    matrix: List = Field(None,
                         deprecated=True,
                         description='Use `dimensions` instead.',
                         exclude=True)

    sampler: Union[LHSSampler, HaltonSampler, SobolSampler,
                   CartesianProduct] = Field(default=LHSSampler(),
                                             discriminator='method',
                                             description=f("""
        For efficient UQ, it may not be necessary to sample the entire matrix
        or hypercube. By default, the cartesian product is taken. For more
        efficient sampling of the space, the following `method` choices are
        available:
        [`latin-hypercube`](en.wikipedia.org/wiki/Latin_hypercube_sampling),
        [`sobol`](en.wikipedia.org/wiki/Sobol_sequence),
        [`halton`](en.wikipedia.org/wiki/Halton_sequence).
        Where `n_samples` gives the number of samples to extract.
        """))

    template: DirectoryPath = Field(
        f'/pfs/work/{getuser()}/jetto/runs/duqtools_template',
        description=f("""
        Template directory to modify. Duqtools copies and updates the settings
        required for the specified system from this directory. This can be a
        directory with a finished run, or one just stored by JAMS (but not yet
        started). By default, duqtools extracts the input IMAS database entry
        from the settings file (e.g. jetto.in) to find the data to modify for
        the UQ runs.
        """))

    template_data: ImasBaseModel = Field(None,
                                         description=f("""
        Specify the location of the template data to modify. This overrides the
        location of the data specified in settings file in the template
        directory.
        """))

    data: DataLocation = Field(DataLocation(),
                               description=f("""
        Where to store the in/output IDS data.
        The data key specifies the machine or imas
        `db` name where to store the data (`db`). duqtools will write the input
        data files for UQ start with the run number given by `run_in_start_at`.
        The data generated by the UQ runs (e.g. from jetto) will be stored
        starting by the run number given by `run_out_start_at`.
        """))

    @validator('matrix', always=True, pre=True)
    def deprecate_matrix(v):
        if v:
            raise DeprecatedValueError(
                "'matrix' has been deprecated, use 'dimensions' instead.")


class SubmitConfigModel(BaseModel):
    """The options of the `submit` subcommand are stored under the `submit` key
    in the config.

    The config describes the commands to start the UQ runs.
    """

    submit_script_name: str = Field(
        '.llcmd', description='Name of the submission script.')
    submit_command: str = Field('sbatch', description='Submission command.')


class StatusConfigModel(BaseModel):
    """The options of the `status` subcommand are stored under the `status` key
    in the config.

    These only need to be changed if the modeling software changes.
    """

    status_file: str = Field('jetto.status',
                             description='Name of the status file.')
    in_file: str = Field('jetto.in',
                         description=f("""
            Name of the modelling input file, will be used to check
            if the subprocess has started.
            """))

    out_file: str = Field('jetto.out',
                          description=f("""
            Name of the modelling output file, will be used to
            check if the software is running.
            """))

    msg_completed: str = Field('Status : Completed successfully',
                               description=f("""
            Parse `status_file` for this message to check for
            completion.
            """))

    msg_failed: str = Field('Status : Failed',
                            description=f("""
            Parse `status_file` for this message to check for
            failures.
            """))

    msg_running: str = Field('Status : Running',
                             description=f("""
            Parse `status_file` for this message to check for
            running status.
            """))


class ConfigModel(BaseModel):
    """The options for the CLI are defined by this model."""
    plot: dict = Field(None,
                       deprecated=True,
                       description='Options are specified via CLI.',
                       exclude=True)

    submit: SubmitConfigModel = Field(
        SubmitConfigModel(),
        description='Configuration for the submit subcommand')
    create: CreateConfigModel = Field(
        CreateConfigModel(),
        description='Configuration for the create subcommand')
    status: StatusConfigModel = Field(
        StatusConfigModel(),
        description='Configuration for the status subcommand')
    workspace: WorkDirectoryModel = WorkDirectoryModel()
    system: Literal['jetto',
                    'dummy'] = Field('jetto',
                                     description='backend system to use')

    @validator('plot', always=True, pre=True)
    def deprecate_matrix(v):
        if v:
            raise DeprecatedValueError(
                "'plot' config has been deprecated, you can remove this "
                'section from your config file.')



import logging
from pathlib import Path
from time import sleep

from .config import cfg
from .jetto import read_namelist
from .models import WorkDirectory

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def has_submit_script(dir: Path) -> bool:
    return (dir / cfg.submit.submit_script_name).exists()


def has_status(dir: Path) -> bool:
    return (dir / cfg.status.status_file).exists()


def is_submitted(dir: Path) -> bool:
    return (dir / 'duqtools.submit.lock').exists()


def status_file_contains(dir: Path, msg) -> bool:
    sf = (dir / cfg.status.status_file)
    with open(sf, 'r') as f:
        content = f.read()
        debug('Checking if content of %s file: %s contains %s', sf, content,
              msg)
        return msg in content


def is_completed(dir: Path) -> bool:
    return status_file_contains(dir, cfg.status.msg_completed)


def is_failed(dir: Path) -> bool:
    return status_file_contains(dir, cfg.status.msg_failed)


def is_running(dir: Path) -> bool:
    return status_file_contains(dir, cfg.status.msg_running)


def completion_percentage(dir: Path) -> int:
    of = (dir / cfg.status.out_file)
    infile = (dir / cfg.status.in_file)
    if not of.exists():
        debug('%s does not exists, but the job is running', of)
        return 0
    if not infile.exists():
        debug('%s does not exists, but the job is running', infile)
        return 0

    with open(of, 'r') as f:
        lines = f.readlines()
        for i in range(len(lines) - 1, 0, -1):
            if 'STEP' in lines[i]:
                time = float(lines[i].split('=')[2].lstrip(' ').split(' ')[0])
                break
        else:
            time = 0.
    nml = read_namelist(infile)
    start = nml['nlist1']['tbeg']
    end = nml['nlist1']['tmax']
    return int(100 * (time - start) / (end - start))


class Status():

    dirs: list
    dirs_submit: list
    dirs_status: list
    dirs_failed: list
    dirs_running: list
    dirs_unknown: list

    def __init__(self):
        debug('Submit config: %s', cfg.submit)

        workspace = WorkDirectory.parse_obj(cfg.workspace)
        runs = workspace.runs

        self.dirs = [Path(run.dirname) for run in runs]
        debug('Case directories: %s', self.dirs)

        debug('Total number of directories: %i', len(self.dirs))
        self.update_status()

    def update_status(self):

        self.dirs_submit = [dir for dir in self.dirs if has_submit_script(dir)]
        self.dirs_status = [dir for dir in self.dirs if has_status(dir)]
        self.dirs_submitted = [dir for dir in self.dirs if is_submitted(dir)]

        self.dirs_completed = [
            dir for dir in self.dirs_status if is_completed(dir)
        ]
        self.dirs_failed = [dir for dir in self.dirs_status if is_failed(dir)]
        self.dirs_running = [
            dir for dir in self.dirs_status if is_running(dir)
        ]

        self.dirs_unknown = [
            dir for dir in self.dirs_status
            if not (dir in self.dirs_completed or dir in self.dirs_failed
                    or dir in self.dirs_running)
        ]

    def simple_status(self):
        """stateless status."""

        info('Total number of directories with submission script : %i',
             len(self.dirs_submit))
        info('      number of not submitted jobs                 : %i',
             (len(self.dirs_submit) - len(self.dirs_status)))
        info('Total number of directories with status     script : %i',
             len(self.dirs_status))
        info('Total number of directories with Completed status  : %i',
             len(self.dirs_completed))
        info('Total number of directories with Failed    status  : %i',
             len(self.dirs_failed))
        info('Total number of directories with Running   status  : %i',
             len(self.dirs_running))
        info('Total number of directories with Unknown   status  : %i',
             len(self.dirs_unknown))

    def progress_status(self):
        """Monitor the directory for status changes."""
        from tqdm import tqdm
        pbar_a = tqdm(total=len(self.dirs), position=0)
        pbar_a.set_description('Submitted jobs            ...')
        pbar_b = tqdm(total=len(self.dirs_submit), position=1)
        pbar_b.set_description('Running jobs              ...')
        pbar_c = tqdm(total=len(self.dirs_submit), position=2)
        pbar_c.set_description('Completed jobs            ...')
        pbar_d = tqdm(total=len(self.dirs_submit), position=3)
        pbar_d.set_description('Failed? jobs              ...')
        while len(self.dirs_completed) < len(self.dirs_submit):
            pbar_a.n = len(self.dirs_submitted)
            pbar_b.n = len(self.dirs_running)
            pbar_c.n = len(self.dirs_completed)
            pbar_d.n = len(self.dirs_failed) + len(self.dirs_unknown)
            pbar_a.refresh()
            pbar_b.refresh()
            pbar_c.refresh()
            pbar_d.refresh()
            sleep(5)
            self.update_status()

    def get_status(self, dir):
        if dir not in self.dirs_submit:
            return 'Unsubmitted '
        if dir in self.dirs_running:
            return 'Running     '
        if dir in self.dirs_failed:
            return 'FAILED      '
        if dir in self.dirs_completed:
            return 'COMPLETED   '
        return 'UNKNOWN'

    def detailed_status(self):
        from tqdm import tqdm
        """detailed_status of all separate runs."""
        pbars = []
        for i, dir in enumerate(self.dirs):
            pbars.append(tqdm(total=100, position=i))
            status = self.get_status(dir)
            pbars[-1].set_description(f'{dir.name:8s}, status: {dir.name:12s}')

        while len(self.dirs_completed) < len(self.dirs_submit):
            for i, dir in enumerate(self.dirs):
                status = self.get_status(dir)
                pbars[i].set_description(
                    f'{dir.name:8s}, status: {status:12s}')
                if dir in self.dirs_running:
                    pbars[i].n = completion_percentage(dir)
                pbars[i].refresh()
            sleep(5)
            self.update_status()


def status(*, progress: bool, detailed: bool, **kwargs):
    """Show status of runs.

    Parameters
    ----------
    progress : bool
        Show progress bar.
    detailed : bool
        Show detailed progress for every job.
    """

    tracker = Status()

    if detailed:
        tracker.detailed_status()
    elif progress:
        tracker.progress_status()
    else:
        tracker.simple_status()



from __future__ import annotations

import logging
import shutil

import click

from .config import cfg
from .ids import ImasHandle
from .models import WorkDirectory
from .operations import confirm_operations, op_queue

logger = logging.getLogger(__name__)


@confirm_operations
def cleanup(out, force, **kwargs):
    """Read runs.yaml and clean the current directory.

    Parameters
    ----------
    out : bool
        Remove output IDS.
    """
    workspace = WorkDirectory.parse_obj(cfg.workspace)

    if workspace.runs_yaml.exists and not force:
        if workspace.runs_yaml_old.exists():
            raise IOError(
                '`runs.yaml.old` exists, use --force to overwrite anyway')

    for run in workspace.runs:
        data_in = ImasHandle.parse_obj(run.data_in)
        data_out = ImasHandle.parse_obj(run.data_out)

        data_in.delete()

        if out:
            data_out.delete()
        else:
            op_queue.add(action=lambda: None,
                         description=click.style('NOT Removing',
                                                 fg='red',
                                                 bold=True),
                         extra_description=f'{data_out}')

        op_queue.add(action=shutil.rmtree,
                     args=(run.dirname, ),
                     description='Removing run dir',
                     extra_description=f'{run.dirname}')

    op_queue.add(action=shutil.move,
                 args=(workspace.runs_yaml, workspace.runs_yaml_old),
                 description='Moving runs.yaml',
                 extra_description=f'{workspace.runs_yaml_old}')



from __future__ import annotations

import logging
import xml.sax
import xml.sax.handler
from getpass import getuser
from typing import TYPE_CHECKING

from packaging import version

from .._logging_utils import LoggingContext
from ..operations import add_to_op_queue
from ._imas import imas

if TYPE_CHECKING:
    from .ids import ImasHandle

PATH_IDSDEF = '/gw/swimas/core/installer/src/3.34.0/ual/4.9.3/xml/IDSDef.xml'


class Parser(xml.sax.handler.ContentHandler):

    def __init__(self):
        xml.sax.handler.ContentHandler.__init__(self)
        self.idss = []

    def startElement(self, name: str, attrs):
        if name == 'IDS':
            ids = {}
            for i in attrs.getNames():
                ids[i] = attrs.getValue(i)
            self.idss.append(ids)

    @classmethod
    def load_idsdef(cls):
        parser = cls()
        xml.sax.parse(PATH_IDSDEF, parser)
        return parser


def get_imas_ual_version():
    """Get imas/ual versions.

    Parsed from a string like: `imas_3_34_0_ual_4_9_3`
    """
    vsplit = imas.names[0].split('_')

    imas_version = version.parse('.'.join(vsplit[1:4]))
    ual_version = version.parse('.'.join(vsplit[5:8]))

    return imas_version, ual_version


@add_to_op_queue('Copy ids from template to', '{target}')
def copy_ids_entry(source: ImasHandle, target: ImasHandle):
    """Copies the ids entry to a new location.

    Parameters
    ----------
    source : ImasHandle
        Source ids entry
    target : ImasHandle
        Target ids entry

    Raises
    ------
    KeyError
        If the IDS entry you are trying to copy does not exist.
    """
    assert target.user == getuser()

    imas_version, _ = get_imas_ual_version()

    idss_in = imas.ids(source.shot, source.run)
    op = idss_in.open_env(source.user, source.db, str(imas_version.major))

    ids_not_found = op[0] < 0
    if ids_not_found:
        raise KeyError('The entry you are trying to copy does not exist')

    idss_out = imas.ids(target.shot, target.run)

    idss_out.create_env(target.user, target.db, str(imas_version.major))
    idx = idss_out.expIdx

    parser = Parser.load_idsdef()

    # Temporarily hide warnings, because this loop is very spammy
    with LoggingContext(level=logging.CRITICAL):

        for ids_info in parser.idss:
            name = ids_info['name']
            maxoccur = int(ids_info['maxoccur'])

            if name in ('ec_launchers', 'numerics', 'sdn'):
                continue

            for i in range(maxoccur + 1):
                ids = idss_in.__dict__[name]
                ids.get(i)
                ids.setExpIdx(idx)  # this line sets the index to the output
                ids.put(i)

    idss_in.close()
    idss_out.close()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import Field
from typing_extensions import Literal

from ..models import AbstractSystem
from ..operations import add_to_op_queue
from ._imas_functions import imas_from_jset_input
from ._jset import JettoSettings

if TYPE_CHECKING:
    from ..ids import ImasHandle
    from ..models import WorkDirectory


class JettoSystem(AbstractSystem):
    """This system implements a wrapper around JETTO, which is part of the
    JINTRAC modelling framework for integrated simulation of Tokamaks.

    For more information:

    - G. Cenacchi, A. Taroni, JETTO: A free-boundary plasma transport code,
        JET-IR (1988)
    - M. Romanelli  2014, Plasma and Fusion research 9, 3403023-3403023
    """
    name: Literal['jetto'] = Field('jetto', description='Name of the system.')

    @staticmethod
    @add_to_op_queue('Writing new batchfile', '{run_name}')
    def write_batchfile(workspace: WorkDirectory, run_name: str):
        from duqtools.jetto import write_batchfile as jetto_write_batchfile
        return jetto_write_batchfile(workspace, run_name)

    @staticmethod
    @add_to_op_queue('Copying template to', '{target_drc}')
    def copy_from_template(source_drc: Path, target_drc: Path):
        from duqtools.jetto import copy_files
        return copy_files(source_drc, target_drc)

    @staticmethod
    def imas_from_path(template_drc: Path):

        jset = JettoSettings.from_directory(template_drc)
        source = imas_from_jset_input(jset)
        assert source.path().exists()
        return source

    @staticmethod
    @add_to_op_queue('Updating imas locations of', '{run}')
    def update_imas_locations(run: Path, inp: ImasHandle, out: ImasHandle):
        jset = JettoSettings.from_directory(run)
        jset_copy = jset.set_imas_locations(inp=inp, out=out)
        jset_copy.to_directory(run)



import sys
from pathlib import Path

import pandas as pd
import streamlit as st

from duqtools._plot_utils import alt_errorband_chart, alt_line_chart
from duqtools.ids import (IDSMapping, ImasHandle, get_ids_tree, rebase_on_ids,
                          rebase_on_time)
from duqtools.ids._io import _get_ids_run_dataframe
from duqtools.utils import read_imas_handles_from_file

try:
    default_workdir = sys.argv[1]
except IndexError:
    default_workdir = str(Path.cwd())

st.title('Plot IDS')

with st.sidebar:
    st.header('Input data')
    work_dir = st.text_input('Work directory', default_workdir)
    data_file = st.text_input('Data file', 'runs.yaml')

inp = Path(work_dir) / data_file

if not inp.exists():
    raise ValueError(f'{inp} does not exist!')

handles = read_imas_handles_from_file(inp)
df = pd.DataFrame.from_dict(
    {index: model.dict()
     for index, model in handles.items()}, orient='index')

with st.expander('IDS data'):
    st.table(df)

prefix = 'profiles_1d/$i'


def ffmt(s):
    return s.replace(prefix + '/', '')


def get_options(a_run):
    a_profile = get_ids_tree(ImasHandle(**a_run), exclude_empty=True)
    return sorted(a_profile.find_by_index(f'{prefix}/.*').keys())


options = get_options(a_run=df.iloc[0])

with st.sidebar:

    default_x_key = options.index(f'{prefix}/grid/rho_tor_norm')
    default_y_val = f'{prefix}/t_i_average'

    x_key = st.selectbox('Select IDS (x)',
                         options,
                         index=default_x_key,
                         format_func=ffmt)

    y_keys = st.multiselect('Select IDS (y)',
                            options,
                            default=default_y_val,
                            format_func=ffmt)

    show_error_bar = st.checkbox(
        'Show errorbar',
        help=(
            'Show standard deviation band around mean y-value. All '
            'y-values are interpolated to put them on a common basis for x.'))


@st.experimental_memo()
def get_data(df, **kwargs):
    """Get and concatanate data for all runs."""
    runs_data = {
        str(name): _get_ids_run_dataframe(ImasHandle(**row), **kwargs)
        for name, row in df.iterrows()
    }

    return pd.concat(runs_data,
                     names=('run',
                            'index')).reset_index('run').reset_index(drop=True)


rebase_on_ids = st.experimental_memo(rebase_on_ids)
rebase_on_time = st.experimental_memo(rebase_on_time)

y_vals = tuple(ffmt(y_key) for y_key in y_keys)
x_val = ffmt(x_key)

for y_val in y_vals:
    st.header(f'{x_val} vs. {y_val}')

    source = get_data(df, keys=(x_val, y_val), prefix='profiles_1d')

    if show_error_bar:
        source = rebase_on_ids(source, base_col=x_val, value_cols=[y_val])
        source = rebase_on_time(source, cols=(x_val, y_val))

        chart = alt_errorband_chart(source, x=x_val, y=y_val)

    else:
        chart = alt_line_chart(source, x=x_val, y=y_val)

    st.altair_chart(chart, use_container_width=True)

with st.form('Save to new IMAS DB entry'):
    a_run = df.iloc[0]

    st.subheader('Template IMAS entry:')

    cols = st.columns(4)

    template = {
        'user': cols[0].text_input('User',
                                   value=a_run.user,
                                   key='user_template'),
        'db': cols[1].text_input('Machine', value=a_run.db, key='db_template'),
        'shot': cols[2].number_input('Shot',
                                     value=a_run.shot,
                                     key='shot_template'),
        'run': cols[3].number_input('Run', value=a_run.run,
                                    key='run_template'),
    }

    template = ImasHandle(**template)

    st.subheader('Target IMAS entry:')

    cols = st.columns(4)

    target = {
        'user':
        cols[0].text_input('User',
                           value=a_run.user,
                           key='user_target',
                           disabled=True),
        'db':
        cols[1].text_input('Machine', value=a_run.db, key='db_target'),
        'shot':
        cols[2].number_input('Shot', value=a_run.shot, key='shot_target'),
        'run':
        cols[3].number_input('Run', step=1, key='run_target'),
    }

    target = ImasHandle(**target)

    submitted = st.form_submit_button('Save')
    if submitted:
        template_data = get_ids_tree(template)

        # pick first time step as basis
        common_basis = template_data[f'profiles_1d/0/{x_val}']

        data = get_data(df, keys=[x_val, *y_vals], prefix='profiles_1d')

        data = rebase_on_ids(data,
                             base_col=x_val,
                             value_cols=y_vals,
                             new_base=common_basis)

        # common_time = [0.0, 0.25, 0.50, 0.75, 1.0]
        common_time = template_data['time']

        # Set to common time basis
        data = rebase_on_time(data,
                              cols=[x_val, *y_vals],
                              new_base=common_time)

        gb = data.groupby(['tstep', x_val])

        agg_funcs = ['mean', 'std']
        agg_dict = {y_val: agg_funcs for y_val in y_vals}

        merged = gb.agg(agg_dict)

        template.copy_ids_entry_to(target)

        core_profiles = target.get('core_profiles')
        ids_mapping = IDSMapping(core_profiles, exclude_empty=False)

        for tstep, group in merged.groupby('tstep'):

            mean = group['t_i_average', 'mean']
            stdev = group['t_i_average', 'std']

            profile = ids_mapping[f'profiles_1d/{tstep}/{y_val}']
            profile[:] = mean

            # This does not work yet, because `*_error_upper` *may* be empty
            # profile_error_upper = ids_mapping[f'profiles_1d/{tstep}/{y_val}_error_upper']
            # profile_error_upper[:] = mean + stdev

        with target.open() as data_entry_target:
            core_profiles.put(db_entry=data_entry_target)

        st.success('Success!')
        st.balloons()



import logging
from pathlib import Path

from .config import Config
from .operations import confirm_operations, op_queue
from .schema import BaseModel

logger = logging.getLogger(__name__)


@confirm_operations
def init(*, config: str, full: bool, force: bool, **kwargs):
    """Initialize a brand new config file with all the default values.

    Parameters
    ----------
    config : str
        Filename of the config.
    full : bool
        Make a config with all the default values
        (otherwise just selected important ones)
    force : bool
        Overwrite config if it already exists.
    **kwargs
        Unused.

    Raises
    ------
    RuntimeError
        When the config already exists.
    """
    cfg = Config()
    BaseModel.__init__(cfg)

    logger.debug(cfg)

    config_filepath = Path(config)

    if config_filepath.exists() and not force:
        raise RuntimeError(
            f'Refusing to overwrite existing CONFIG, {config_filepath}, '
            'use --force if you really want to')

    logger.debug('Creating default cfg.yaml')

    if full:
        cfg_yaml = cfg.yaml()
    else:
        cfg_yaml = cfg.yaml(
            include={
                'workspace': True,
                'create': {'dimensions', 'sampler', 'template', 'data'},
            })

    op_queue.add(action=lambda: open(config_filepath, 'w').write(cfg_yaml),
                 description='Writing out',
                 extra_description=f'{config_filepath} config file')



import logging
import subprocess
from pathlib import Path
from typing import Any, List

from .config import cfg
from .models import WorkDirectory
from .operations import add_to_op_queue, confirm_operations

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


@add_to_op_queue('Submitting', '{run_dir}')
def submit_job(lockfile, cmd, run_dir):
    debug(f'Put lockfile in place for {lockfile}')
    lockfile.touch()

    info(f'submitting script {cmd}')
    ret = subprocess.run(cmd, check=True, capture_output=True)
    info(f'submission returned: {ret.stdout}')
    with open(lockfile, 'wb') as f:
        f.write(ret.stdout)


@confirm_operations
def submit(*, force: bool, **kwargs):
    """submit. Function which implements the functionality to submit jobs to
    the cluster.

    Parameters
    ----------
    force : bool
        force the submission even in the presence of lockfiles
    """

    if not cfg.submit:
        raise Exception('submit field required in config file')

    debug('Submit config: %s', cfg.submit)

    workspace = WorkDirectory.parse_obj(cfg.workspace)
    runs = workspace.runs

    run_dirs = [Path(run.dirname) for run in runs]
    debug('Case directories: %s', run_dirs)

    for run_dir in run_dirs:
        submission_script = run_dir / cfg.submit.submit_script_name
        if submission_script.is_file():
            info('Found submission script: %s ; Ready for submission',
                 submission_script)
        else:
            info('Did not found submission script %s ; Skipping directory...',
                 submission_script)
            continue

        status_file = run_dir / cfg.status.status_file
        if status_file.exists() and not force:
            if not status_file.is_file():
                info('Status file %s is not a file', status_file)
            with open(status_file, 'r') as f:
                info('Status of %s: %s. To rerun enable the --force flag',
                     status_file, f.read())
            info('Skipping directory %s ; due to existing status file',
                 run_dir)
            continue

        lockfile = run_dir / 'duqtools.submit.lock'
        if lockfile.exists() and not force:
            info(
                'Skipping %s, lockfile exists, enable --force to submit again',
                run_dir)
            continue

        submit_cmd = cfg.submit.submit_command.split()
        cmd: List[Any] = [*submit_cmd, submission_script]

        submit_job(lockfile, cmd, run_dir)



from ._basemodel import BaseModel
from ._dimensions import ARange, IDSOperation, IDSOperationDim, LinSpace
from ._imas import ImasBaseModel

__all__ = [
    'BaseModel',
    'IDSOperation',
    'IDSOperationDim',
    'ImasBaseModel',
    'LinSpace',
    'ARange',
]



"""Functions to interface with `jetto.jset` files."""
from __future__ import annotations

from copy import deepcopy
from logging import debug
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, TextIO, Tuple

if TYPE_CHECKING:
    from .._types import PathLike
    from ..ids import ImasHandle

DEFAULT_FILENAME = 'jetto.jset'

HEADER = """!================================================================
!                      JETTO SETTINGS FILE
!================================================================"""

JINTRAC_CONFIG_VARS = (
    {
        'name': 'shot_in',
        'type': int,
        'key': 'SetUpPanel.idsIMASDBShot',
        'doc': 'Input IDS shot'
    },
    {
        'name': 'shot_out',
        'type': int,
        'key': 'SetUpPanel.shotNum',
        'doc': 'Output IDS shot'
    },
    {
        'name': 'run_in',
        'type': int,
        'key': 'SetUpPanel.idsIMASDBRunid',
        'doc': 'Input IDS run'
    },
    {
        'name': 'run_out',
        'type': int,
        'key': 'JobProcessingPanel.idsRunid',
        'doc': 'Output IDS run'
    },
    {
        'name': 'user_in',
        'type': str,
        'key': 'SetUpPanel.idsIMASDBUser',
        'doc': 'Input IDS user'
    },
    {
        'name': 'machine_in',
        'type': str,
        'key': 'SetUpPanel.idsIMASDBMachine',
        'doc': 'Input IDS machine'
    },
    {
        'name': 'machine_out',
        'type': str,
        'key': 'SetUpPanel.machine',
        'doc': 'Output IDS machine'
    },
    {
        'name': 'noprocessors',
        'type': str,
        'key': 'JobProcessingPanel.numProcessors',
        'doc': 'JobProcessingPanel.numProcessors'
    },
    {
        'name': 'tstart',
        'type': float,
        'key': 'SetUpPanel.startTime',
        'doc': 'Start time'
    },
    {
        'name': 'tend',
        'type': float,
        'key': 'SetUpPanel.endTime',
        'doc': 'End time'
    },
    {
        'name': 'run_dir',
        'type': str,
        'key': 'AppPanel.openPrvSetDir',
        'doc': 'Location of the run directory'
    },
    {
        'name': 'run_dir_name',
        'type': str,
        'key': 'JobProcessingPanel.runDirNumber',
        'doc': 'Name of the run directory'
    },
)


def parse_section(section: List[str]) -> Tuple[str, Dict[str, str]]:
    """Parse section of settings file.

    Parameters
    ----------
    section : List[str]
        Section in the jetto settings file

    Returns
    -------
    Tuple[str, Dict[str, str]]
        Title and dictionary with settings
    """
    title, *section = section
    title = title.strip('*')

    data = {}
    for line in section:
        key, value = line.split(':', 1)
        data[key.strip()] = value.strip()

    return title, data


def parse_jset(file: TextIO) -> Dict[str, Dict[str, str]]:
    """Parse jetto settings file.

    Parameters
    ----------
    file : TextIO
        Open file or text buffer

    Returns
    -------
    sections : Dict[str, Dict[str, str]]
        Return nested dictionary with settings from jset file.
    """
    sections = []

    read_block = False
    block: List[str] = []

    for line in file:
        line = line.strip()

        if not line:
            continue
        elif line == '*':
            read_block = True
            if block:
                sections.append(parse_section(block))
            block = []
        elif line.lower() == '*eof':
            break
        elif read_block:
            block.append(line)

    return dict(sections)


def read_jset(path: PathLike) -> Dict[str, Dict[str, str]]:
    """Read a jetto settings file saved by JAMS.

    Parameters
    ----------
    path : PathLike
        Path to jset file.

    Returns
    -------
    settings : Dict[str, Dict[str, str]]
        Return nested dictionary with settings from jset file.
    """
    with open(path) as f:
        settings = parse_jset(f)

    return settings


def write_jset(path: PathLike, settings: Dict[str, Dict[str, str]]):
    """Write a jetto settings file.

    Parameters
    ----------
    path : PathLike
        Path to which the settings are saved.
    settings : Dict[str, Dict[str, str]]
        Jetto settings dictionary.
    """

    def _line_gen():
        yield HEADER

        for title, section in settings.items():
            yield '*'
            yield f'*{title}'
            for key, value in section.items():
                yield f'{key:<61}: {value}'

        yield from ('*', '*EOF', '')

    lines = (f'\n{line}' for line in _line_gen())

    with open(path, 'w') as f:
        f.writelines(lines)


class JettoSettings:

    def __init__(self, mapping: Dict[str, Dict[str, str]]):
        self.raw_mapping = mapping

    @property
    def metadata(self):
        return self.raw_mapping['File Details']

    @property
    def settings(self):
        return self.raw_mapping['Settings']

    def __new__(cls, *args, **kwargs):

        def setter(jset_key: str, jset_type):

            def f(self, value):
                self.settings[jset_key] = str(value)

            return f

        def getter(jset_key: str, jset_type):

            def f(self):
                return jset_type(self.settings[jset_key])

            return f

        for variable in JINTRAC_CONFIG_VARS:
            jset_key = variable['key']
            jset_type = variable['type']
            name = variable['name']
            doc = variable['doc']
            prop = property(
                fget=getter(jset_key, jset_type),
                fset=setter(jset_key, jset_type),
                doc=doc,
            )
            setattr(cls, name, prop)

        return super().__new__(cls)

    def copy(self):
        """Return a copy of this instance."""
        return deepcopy(self)

    @property
    def components(self):
        """Active components e.g. EDGE2D, JETTO, HCD (uppercase)"""
        components = ['JETTO']
        if self.settings['JobProcessingPanel.selIdsRunid']:
            components.append('IDSOUT')
        if self.settings['ExternalWFPanel.select']:
            components.append('HCD')
        if self.settings['SetUpPanel.selReadIds']:
            components.append('IDSIN')
        return components

    def to_directory(self, directory: PathLike):
        """Write a new jetto.jset to the given directory.

        Parameters
        ----------
        directory : PathLike
            Name of output directory
        """
        directory = Path(directory)

        self.run_dir = str(directory)
        self.run_dir_name = directory.name

        filename = directory / DEFAULT_FILENAME
        write_jset(filename, self.raw_mapping)
        debug('write %s', filename)

    @classmethod
    def from_file(cls, path: PathLike) -> JettoSettings:
        """Read JettoSettings from 'jetto.jset' file.

        Parameters
        ----------
        path : PathLike
            Path to `jetto.jset`

        Returns
        -------
        jset : JettoSettings
            Instance of `JettoSettings`
        """
        mapping = read_jset(path)
        jset = cls(mapping)
        return jset

    @classmethod
    def from_directory(cls, path: PathLike) -> JettoSettings:
        """Read JettoSettings from 'jetto.jset' file in given directory.

        Parameters
        ----------
        path : PathLike
            Path to directory containing `jetto.jset`

        Returns
        -------
        jset : JettoSettings
            Instance of `JettoSettings`
        """
        filename = Path(path) / DEFAULT_FILENAME
        return cls.from_file(filename)

    def copy_and_patch(self,
                       settings: dict = None,
                       metadata: dict = None) -> JettoSettings:
        """Patch a copy of the jetto settings.

        Parameters
        ----------
        settings : dict, optional
            Settings to update
        metadata : dict, optional
            Metadata to update

        Returns
        -------
        jset_copy : JettoSettings
            Patched copy.
        """
        jset_copy = self.copy()

        if settings:
            jset_copy.settings.update(settings)
        if metadata:
            jset_copy.metadata.update(metadata)

        return jset_copy

    def set_imas_locations(self, inp: ImasHandle,
                           out: ImasHandle) -> JettoSettings:
        """Make a copy with updated IDS locations for input / output.

        Parameters
        ----------
        inp : ImasHandle
            IMAS description of where the input data is stored.
        out : ImasHandle
            IMAS description of where the output data should be stored.

        Returns
        -------
        jset_new : JettoSettings
            Copy of the jetto settings with updated IDS locations.
        """
        jset_new = self.copy()

        jset_new.user_in = inp.user
        jset_new.machine_in = inp.db
        jset_new.shot_in = inp.shot
        jset_new.run_in = inp.run

        jset_new.machine_out = out.db
        jset_new.shot_out = out.shot
        jset_new.run_out = out.run

        return jset_new
